"""
The class CompressMolecules has a function Compress that takes a dilute
system of molecules and slowly compresses it to the desired density while
running MD. Currently only one bond-type is supported
"""

import sys
import string
import numpy

from rumd import *
from rumdSimulation import rumdSimulation


class CompressMolecules:
    def __init__(self):
        self.params = {
            "desiredRho": 0.9,
            "r_bond": 0.8,
            "k_bond": 200.0,
            "config_filename": "start.xyz.gz",
            "topology_filename": "start.top",
            "T": 0.5,
            "dt":  0.004,
            "numIterations": 100,
            "nMinorTimeSteps": 1000,
            "n_equil_steps":100000
            }
        
    def Compress(self):

        P = self.params

        sim = rumdSimulation(P["config_filename"])

        # Create LJ pair-potential
        potential = Pot_LJ_12_6()
        # Set up parameters for Lennard-Jones interactions
        # (keep all interactions the same--assuming at most two types)
        # i  j   Sig   Eps  Rcut (in units of Sigma_ij)
        potential.SetParams(0, 0, 1.00, 1.00, 2.5)
        potential.SetParams(0, 1, 1.00, 1.00, 2.5)
        potential.SetParams(1, 0, 1.00, 1.00, 2.5)
        potential.SetParams(1, 1, 1.00, 1.00, 2.5)
        sim.SetPotential(potential)

        # integrator
        itg = IntegratorNVT(timeStep=P["dt"], targetTemperature=P["T"])
        sim.SetIntegrator(itg);

        # Molecule stuff
        sim.ReadMoleculeData(P["topology_filename"]) 
        sim.SetBondHarmonic(0, P["r_bond"], P["k_bond"])
        nParticles = sim.sample.GetNumberOfParticles()
        vol0 = sim.sample.GetSimulationBox().GetVolume()
        rho0 = nParticles/vol0

        scaleFactor = numpy.exp(-numpy.log(P["desiredRho"]/rho0)/3/P["numIterations"])
        print "rho0",rho0
        print "scaleFactor",scaleFactor

        for majorIdx in xrange(P["numIterations"]):
            vol = sim.sample.GetSimulationBox().GetVolume()
            print "density:",nParticles/vol
            sim.Run(P["nMinorTimeSteps"], suppressAllOutput=True)

            sim.sample.IsotropicScaleSystem(scaleFactor)

        print "Equilibrating at final density"
        sim.SetOutputParameters("energies","linear",interval=50)
        #sim.SetOutputParameters("trajectory","linear",interval=500)
        #sim.SetBlockSize(10000)
        sim.Run(P["n_equil_steps"])
        print "Writing final configuration to equil.xyz.gz"
        sim.sample.WriteConf("equil.xyz.gz")

if __name__ == "__main__":

    em = CompressMolecules()

    if "--help" in sys.argv:
        print "Usage: python CompressMolecules.py <a=b> ... "
        print "Arguments and their default values are as follows:"
        for k in em.params.keys():
            print k, em.params[k]
        sys.exit(0)

    for item in sys.argv[1:]:
        key, val = string.split(item,"=")
        em.params[key] = string.atof(val)
        
    em.Compress()
