#include "rumd_msd.h"
#include <iostream>
#include <sstream>
#include <iomanip>
#include <fstream>

rumd_msd::rumd_msd() : directory("TrajectoryFiles"),
		       C0(),
		       C1(),
		       Time(0),
		       Count(0),
		       R2(0),
		       R4(0),
		       Fs(0),
		       Fs2(0),  
		       R2cm(0),
		       R4cm(0),
		       Fscm(0),
		       Fs_(0),
		       qvalues(0),
		       numParticlesOfType(0),
		       MaxDataIndex(0),
		       num_types(0),
		       num_particles(0),
		       particlesPerMol(1),
		       num_moleculetypes(0),
		       verbose(false),
		       numFilenameDigits(4),
		       dt(0.0) {}



void rumd_msd::AllocateArrays(unsigned int set_MaxDataIndex, unsigned int set_num_types) {
  if( (set_MaxDataIndex != MaxDataIndex || set_num_types != num_types) && MaxDataIndex > 0 && num_types > 0)
    FreeArrays();

  MaxDataIndex = set_MaxDataIndex;
  num_types = set_num_types;

  Time = new unsigned long int[MaxDataIndex];
  R2    = new double*[MaxDataIndex];
  R4    = new double*[MaxDataIndex];
  Fs    = new double*[MaxDataIndex];
  Fs2   = new double*[MaxDataIndex];

  Count = new unsigned long int[MaxDataIndex];
  for (unsigned int i=0; i<MaxDataIndex; i++ ) {
    R2[i]    = new double[num_types];
    R4[i]    = new double[num_types];
    Fs[i]    = new double[num_types];
    Fs2[i]   = new double[num_types];
  } // loop over i

  qvalues = new double[num_types];
  numParticlesOfType = new int[num_types];
  Fs_ = new double[num_types];
  
}

void rumd_msd::AllocateArraysCM( unsigned int ppm ) {
  if(MaxDataIndex == 0)
    throw RUMD_Error("rumd_msd","AllocateArraysCM","MaxDataIndex has not been set---must call AllocateArrays before AllocateArraysCM");
  
  particlesPerMol = ppm;
  
  unsigned int set_num_moleculetypes = particlesPerMol  > 1 ? 1 : 0;
  // For now only deal with one type of molecules at most 

  if(set_num_moleculetypes != num_moleculetypes && num_moleculetypes > 0)
    FreeArraysCM();


  num_moleculetypes = set_num_moleculetypes;

  R2cm  = new double*[MaxDataIndex];
  R4cm  = new double*[MaxDataIndex];
  Fscm  = new double*[MaxDataIndex];
  
  for (unsigned int i=0; i<MaxDataIndex; i++ ) {
    R2cm[i]  = new double[num_moleculetypes];
    R4cm[i]  = new double[num_moleculetypes];
    Fscm[i]  = new double[num_moleculetypes];
  }
  
}
  

void rumd_msd::ResetArrays() {
  if(MaxDataIndex == 0)
    throw RUMD_Error("rumd_msd","ResetArrays","Called Reset before arrays have been allocated");

  for (unsigned int i=0; i<MaxDataIndex; i++ ) {
    Count[i] = 0;

    for (unsigned int j=0; j<num_types; j++) {
      R2[i][j] = 0.0;
      R4[i][j] = 0.0;
      Fs[i][j] = 0.0;
      Fs2[i][j] = 0.0;
    }

    if(num_moleculetypes > 0)
      for (unsigned int j=0; j<num_moleculetypes; j++) {
	R2cm[i][j] = 0.0;
	R4cm[i][j] = 0.0;
	Fscm[i][j] = 0.0;
      }
  } // loop over i
  
  memset(numParticlesOfType, 0, num_types*sizeof(int));

}


void rumd_msd::FreeArrays() {
  for (unsigned int i=0; i<MaxDataIndex; i++ ) {
    delete R2[i];
    delete R4[i];
    delete Fs[i];
    delete Fs2[i];
  }

  delete Time;
  delete R2;
  delete R4;
  delete Fs;
  delete Fs2;

  delete Count;
  delete qvalues;
  delete numParticlesOfType;

  delete Fs_;
}

void rumd_msd::FreeArraysCM() {
  for (unsigned int i=0; i<MaxDataIndex; i++ ) {
    delete R2cm[i];
    delete R4cm[i];
    delete Fscm[i];
  }
  
  delete R2cm;
  delete R4cm;
  delete Fscm;
  
}

void rumd_msd::CalcR2(Conf *C0, Conf *Ct, double *R2,  double *R4, double *Fs, double *Fs2, unsigned int num_types, double *q) {
  float L = C0->metaData.L[0];

  //double *Fs_;
  //Fs_ = new double[num_types];

  for (unsigned int i=0; i<num_types; i++) Fs_[i] = 0.0;

  for (unsigned int i=0; i<num_particles; i++) {
    unsigned int Type = C0->P[i].MyType;
    assert(Type<num_types);
    assert(Type==Ct->P[i].MyType);
    double Rsq = C0->P[i].Rsq(&Ct->P[i], L, L, L );
    R2[Type]  += Rsq;
    R4[Type]  += Rsq*Rsq;
    Fs_[Type] += C0->P[i].Fsq(&Ct->P[i], L, L, L, q[Type]);
  }
  for (unsigned int i=0; i<num_types; i++) {
    Fs[i] += Fs_[i];
    Fs2[i] += Fs_[i]*Fs_[i];
  }
  //delete[] Fs_;

}

void rumd_msd::CalcR2cm(Conf *C0, Conf *Ct, double *R2,  double *R4, double *Fs, unsigned int num_types, double *q) {
  // Calculate dynamical properties of center of mass of molecules, assuming they all have 'particlesPerMol' atoms

  double Lx = C0->metaData.L[0];
  double Ly = C0->metaData.L[1];
  double Lz = C0->metaData.L[2];

  for (unsigned int i=0; i<num_particles; i+=particlesPerMol) {
    double dXtot   = 0.0;
    double dYtot   = 0.0;
    double dZtot   = 0.0;
    double totMass = 0.0;
   
    for (unsigned int j=0; j<particlesPerMol; j++) {
      unsigned int myType = C0->P[i+j].MyType;
      assert(myType<num_types);
      assert(myType==Ct->P[i+j].MyType);
      double myMass = C0->metaData.massOfType[myType];
      totMass += myMass;
 
      // Calculate the displacement of the i+j'th particle
      double dx = Ct->P[i+j].x - C0->P[i+j].x + Lx*(Ct->P[i+j].Imx - C0->P[i+j].Imx);
      double dy = Ct->P[i+j].y - C0->P[i+j].y + Ly*(Ct->P[i+j].Imy - C0->P[i+j].Imy);
      double dz = Ct->P[i+j].z - C0->P[i+j].z + Lz*(Ct->P[i+j].Imz - C0->P[i+j].Imz);
           
      // Add to the center of mass discplacement
      dXtot += myMass*dx;
      dYtot += myMass*dy;
      dZtot += myMass*dz;
    }                         
    
    dXtot /= totMass;
    dYtot /= totMass;
    dZtot /= totMass;
    
    double Rsq = dXtot*dXtot + dYtot*dYtot + dZtot*dZtot;
                    
    // We are not (yet) differentiating between different types of molecules
    R2[0] += Rsq;
    R4[0] += Rsq*Rsq;
    Fs[0] += (cos(dXtot*q[0]) + cos(dYtot*q[0]) + cos(dZtot*q[0]) )/3.0;
  }
}


void rumd_msd::ComputeAll(unsigned int first_block, int last_block,  unsigned int particlesPerMol) {
  unsigned int last_saved_block;
  unsigned int last_block_to_read;
 
  Conf *C0 = new Conf();
  Conf *Ct = new Conf();

  gzFile gzfp;
  FILE *fp;

  std::string lastComp_filename = directory+"/LastComplete_block.txt";

  fp=fopen(lastComp_filename.c_str(),"r");
  if (!fp)
    throw RUMD_Error("rumd_msd","ComputeAll", std::string("Error opening file ") + lastComp_filename + ":" + strerror(errno));

  fscanf(fp, "%u", &last_saved_block);
  fclose(fp);

  if(first_block > last_saved_block) {
    std::cerr << "Warning: user-specified first_block out of range; setting to zero" << std::endl;
    first_block = 0;
  }
  if(last_block < 0)
    last_block_to_read = last_saved_block;
  else
    if((unsigned int)last_block > last_saved_block) {
      last_block_to_read = last_saved_block;
      std::cerr << "Warning, specified last_block out of range; using last saved block" << std::endl;
    }
    else
      last_block_to_read = (unsigned int) last_block;

  if(first_block > last_block_to_read)
    std::cerr << "Warning: last_block smaller than first_block: no rdfs will be computed" << std::endl;

  // Read metaData from first configuration in first block
  std::ostringstream C0Name;
  C0Name << directory << "/block" << std::setfill('0') << std::setw(numFilenameDigits) << std::setprecision(numFilenameDigits) << first_block << ".xyz.gz";
  
  if(verbose)
    std::cout << "Reading block " << first_block << ": " << C0Name << std::endl << std::endl;

  gzfp=gzopen(C0Name.str().c_str(),"r");
  if (!gzfp)
    throw RUMD_Error("rumd_msd","ComputeAll",std::string("Error opening file") + C0Name.str() + ": " + strerror(errno) );

  C0->Read(gzfp,C0Name.str().c_str(),verbose);
  gzclose(gzfp);

  // Store the metaData we need
  // float L  = C0->metaData.L[0];     // Assuming Lx=Ly=Lz for now...
  dt = C0->metaData.dt;
  num_particles = C0->num_part;
  unsigned int set_num_types = C0->metaData.numTypes;
  unsigned int MaxIndex = C0->metaData.logLin.maxIndex;
  unsigned int MaxInterval = C0->metaData.logLin.maxInterval;
  if(MaxInterval)
    throw RUMD_Error("rumd_msd","ComputeAll","Only logarithmically saved data allowed");

  int set_MaxDataIndex = MaxIndex + int(log2(last_block_to_read-first_block+1)+1);

  AllocateArrays(set_MaxDataIndex, set_num_types);

  AllocateArraysCM(particlesPerMol);

  ResetArrays();


  fp=fopen("qvalues.dat","r");
  if (!fp){
    fprintf (stderr, "Error opening file qvalues.dat: %s\n", strerror(errno));
    exit(EXIT_FAILURE);
  }

  std::cout << "\nReading q-values: ";
  for (unsigned int i=0; i<num_types; i++) {
    fscanf(fp, "%lf", &qvalues[i]);
    std::cout << qvalues[i] << " ";
    assert(qvalues[i]>0);
   }
  std::cout << std::endl << std::endl;
  fclose(fp);

  // Read and print the number of particles of each type
  for (unsigned int i=0; i<num_particles; i++) {
    unsigned int Type = C0->P[i].MyType;
    assert(Type<num_types);
    numParticlesOfType[Type]++;
  }

  for (unsigned int i=0; i<num_types; i++) {
    std::cout << "Particles of type " << i << ": " << numParticlesOfType[i] << std::endl;
  }


  std::cout << "\nReading block: ";
  for (unsigned int i=first_block; i<=last_block_to_read; i++) {

    // Read first configuration of block 'i'
    std::ostringstream C0Name;
    C0Name << directory << "/block" << std::setfill('0') << std::setw(numFilenameDigits) << std::setprecision(numFilenameDigits) << i << ".xyz.gz";


    //char *C0Name;
    //asprintf(&C0Name,(directory + "/block%.4u.xyz.gz").c_str(),i);
    if ((i-first_block)%10==0) std::cout << i << " ";
    fflush(stdout);
    gzfp=gzopen(C0Name.str().c_str(),"r");
    C0->Read(gzfp,C0Name.str().c_str(),0);

    // Check that relevant metaData didn't change
    assert(C0->metaData.dt       == dt);
    assert(C0->num_part          == num_particles);
    assert(C0->metaData.numTypes == num_types);
    assert(MaxIndex == C0->metaData.logLin.maxIndex);

    // Analyze configurations in the same block-file
    for (unsigned int j=1; j<=MaxIndex; j++) {
      Ct->Read(gzfp,C0Name.str().c_str(),0);
      Count[j-1]++;
      CalcR2(C0, Ct, R2[j-1], R4[j-1], Fs[j-1], Fs2[j-1], num_types, qvalues);
      if (num_moleculetypes>0) {
        CalcR2cm(C0, Ct, R2cm[j-1], R4cm[j-1], Fscm[j-1], num_types, qvalues);
      }
    }
    gzclose(gzfp);
 
    // Analyze configurations in the later block-files
    for (unsigned int j=1; i+(1<<j)<=last_block_to_read; j++) {
      std::ostringstream C0Name;
      C0Name << directory << "/block" << std::setfill('0') << std::setw(numFilenameDigits) << std::setprecision(numFilenameDigits) << i+(1<<j) << ".xyz.gz";
      gzfp=gzopen(C0Name.str().c_str(),"r");
      Ct->Read(gzfp,C0Name.str().c_str(),0);
      unsigned int jj = MaxIndex+j-1;
      assert(jj<MaxDataIndex);
      Count[jj]++;
      CalcR2(C0, Ct, R2[jj], R4[jj], Fs[jj], Fs2[jj], num_types, qvalues);
      if (num_moleculetypes>0) {
        CalcR2cm(C0, Ct, R2cm[jj], R4cm[jj], Fscm[jj], num_types, qvalues);
      }
      gzclose(gzfp);
     }
  }
  std::cout << std::endl;
}

///////////////// HELPER FUNCTIONS FOR PYTHON INTERFACE /////////////////

void rumd_msd::Copy_MSD_To_Array(double (*msd_array)[2], unsigned int type) {
  if (MaxDataIndex == 0)
    throw RUMD_Error("rumd_msd","Copy_MSD_To_Array","No data! Call ComputeAll first");
  if(type >= num_types)
    throw RUMD_Error("rumd_msd","Copy_MSD_To_Array","Invalid type index");

  for (unsigned int i=0; i<MaxDataIndex; i++ ) {
    msd_array[i][0] = (1UL<<i)*dt;
    if (Count[i]>0)
      msd_array[i][1] = R2[i][type]/float(Count[i]*numParticlesOfType[type]);
    else
      msd_array[i][1] = 0.0;
  } 
}


void rumd_msd::Copy_ISF_To_Array(double (*isf_array)[2], unsigned int type) {
  if (MaxDataIndex == 0)
    throw RUMD_Error("rumd_msd","Copy_ISF_To_Array","No data! Call ComputeAll first");
  if(type >= num_types)
    throw RUMD_Error("rumd_msd","Copy_ISF_To_Array","Invalid type index");

  for (unsigned int i=0; i<MaxDataIndex; i++ ) {
    isf_array[i][0] = (1UL<<i)*dt;
    if (Count[i]>0)
      isf_array[i][1] = Fs[i][type]/float(Count[i]*numParticlesOfType[type]);
    else
    isf_array[i][1] = 0.0;
  } 
}

void rumd_msd::Copy_Alpha2_To_Array(double (*alpha2_array)[2], unsigned int type) {
  if (MaxDataIndex == 0)
    throw RUMD_Error("rumd_msd","Copy_Alpha2_To_Array","No data! Call ComputeAll first");
  if(type >= num_types)
    throw RUMD_Error("rumd_msd","Copy_Alpha2_To_Array","Invalid type index");

  for (unsigned int i=0; i<MaxDataIndex; i++ ) {
    alpha2_array[i][0] = (1UL<<i)*dt;
    if (Count[i]>0)
      alpha2_array[i][1] = 3.f/5.f*R4[i][type]/powf(R2[i][type],2.f)*float(Count[i]*numParticlesOfType[type])-1.f;
    else
      alpha2_array[i][1] = 0.0;
  } 
}

////////////// FILE OUTPUT FUNCTIONS /////////////////////////////////////

void rumd_msd::WriteMSD(const std::string& filename) {
  std::ofstream outfile(filename.c_str());

  for (unsigned int i=0; i<MaxDataIndex; i++ ) {
    if (Count[i]>0) {
      outfile << (1UL<<i)*dt;
      for (unsigned int j=0; j<num_types; j++) {
	outfile << " " << R2[i][j]/float(Count[i]*numParticlesOfType[j]);
      }
      outfile << std::endl;
    }
  }
  outfile.close();
}


void rumd_msd::WriteMSD_CM(const std::string& filename) {
  std::ofstream outfile(filename.c_str());

  for (unsigned int i=0; i<MaxDataIndex; i++ ) {
    if (Count[i]>0) {
      outfile << (1UL<<i)*dt;
      for (unsigned int j=0; j<num_moleculetypes; j++) {
	outfile << " " << R2cm[i][j]/float(Count[i]*num_particles)*float(particlesPerMol);
      }
      outfile << std::endl;
    }
  }
  outfile.close();
}


void rumd_msd::WriteAlpha2(const std::string& filename) {
  std::ofstream outfile(filename.c_str());

  for (unsigned int i=0; i<MaxDataIndex; i++ ) {
    if (Count[i]>0) {
      outfile << (1UL<<i)*dt << " ";
      for (unsigned int j=0; j<num_types; j++) {
	outfile << " " << 3.f/5.f*R4[i][j]/powf(R2[i][j],2.f)*float(Count[i]*numParticlesOfType[j])-1.f;
      }
      outfile << std::endl;
    }
  }
}

void rumd_msd::WriteFs(const std::string& filename) {
  std::ofstream outfile(filename.c_str());

  for (unsigned int i=0; i<MaxDataIndex; i++ ) {
    if (Count[i]>0) {
      outfile << (1UL<<i)*dt << " ";
      
      for (unsigned int j=0; j<num_types; j++) {
	outfile << " " << Fs[i][j]/float(Count[i]*numParticlesOfType[j]);
      }
      outfile << std::endl;
    }
  }
  outfile.close();
}
 

void rumd_msd::WriteFS_CM(const std::string& filename) {
  std::ofstream outfile(filename.c_str());

  for (unsigned int i=0; i<MaxDataIndex; i++ ) {
    if (Count[i]>0) {
      outfile << (1UL<<i)*dt << " ";
      for (unsigned int j=0; j<num_moleculetypes; j++) {
	outfile << " " << Fscm[i][j]/float(Count[i]*num_particles)*float(particlesPerMol);
      }
      outfile << std::endl;
    }
  }
  outfile.close();
}


void rumd_msd::WriteFS_SQ(const std::string& filename) {
  std::ofstream outfile(filename.c_str());

  for (unsigned int i=0; i<MaxDataIndex; i++ ) {
    if (Count[i]>0) {
      outfile << (1UL<<i)*dt << " ";
      for (unsigned int j=0; j<num_types; j++) {
	outfile << " " << Fs2[i][j]/float(Count[i]*numParticlesOfType[j]);
      }
      outfile << std::endl;
    }
  }
  outfile.close();
}
