

/*
  Copyright (C) 2010  Thomas Schrøder
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  LICENSE.txt file for license details.
*/

#include "rumd/DihedralPotential.h"
#include "rumd/MoleculeData.h"
#include "rumd/Sample.h"




DihedralRyckaert::DihedralRyckaert(MoleculeData *globalM, Sample *globalS){

  M = globalM;
  S = globalS;
  SetID_String("DihedralRyckaert");
  
  num_threads = 16;
  num_blocks = M->num_angles/num_threads + 1;

}

__device__ inline void DihedralAtomicFloatAdd(float *address, float val){
  int i_val = __float_as_int(val);
  int tmp0 = 0;
  int tmp1;
  
  while( (tmp1 = atomicCAS((int *)address, tmp0, i_val)) != tmp0){
    tmp0 = tmp1;
    i_val = __float_as_int(val + __int_as_float(tmp1));
  }
}

__global__ void DihedralResetForces( float4 *f, float4 *w ){

  f[blockIdx.x].x = f[blockIdx.x].y = f[blockIdx.x].z = f[blockIdx.x].w = 0.0;
  w[blockIdx.x].w = 0.0;

}

void DihedralRyckaert::CalcF(bool initialize, bool calc_stresses){

  unsigned num_dihedrals = M->GetNumberOfDihedrals();
  
  if ( initialize )
    DihedralResetForces<<<S->GetNumberOfParticles(),1>>>(S->particleData.d_f, 
							 S->particleData.d_w);


  Ryckaert<1><<<num_blocks, num_threads>>>( S->particleData.d_r, S->particleData.d_f, 
					    M->SB, M->SB->GetDevicePointer(), 
					    M->d_dlist, M->d_dtype, M->d_dplist, 
					    M->d_epot_dihedral, M->d_dihedrals, 
					    num_threads, num_dihedrals );

}

double DihedralRyckaert::GetPotentialEnergy(){

  unsigned num_dihedrals = M->GetNumberOfDihedrals();

  
  Ryckaert<2><<<num_blocks, num_threads>>>( S->particleData.d_r, S->particleData.d_f, 
					    M->SB, M->SB->GetDevicePointer(), 
					    M->d_dlist, M->d_dtype, M->d_dplist, 
					    M->d_epot_dihedral, M->d_dihedrals, 
					    num_threads, num_dihedrals );
  
  size_t nbytes = num_dihedrals*sizeof(float);
  cudaMemcpy(M->h_epot_dihedral, M->d_epot_dihedral, nbytes, cudaMemcpyDeviceToHost);
  
  double epot = 0.0;
  for ( unsigned n=0; n<num_dihedrals; n++ )  
    epot += (double)M->h_epot_dihedral[n];

  return epot;
}


template <int energy, class Simbox> 
__global__ void Ryckaert( float4 *r, float4 *f,  Simbox *simbox, float *simboxpointer, 
			  uint4 *dlist, uint1 *dtype, float *plist, 
			  float *d_epot_dihedral, float *d_dihedrals,
			  int num_threads, int num_dihedrals ){

  // First approach is to let one thread work on a dihedral
  float4 dr1, dr2, dr3;    
  float f1, f2, p[6];

  
  unsigned dihedral_index = blockIdx.x*num_threads + threadIdx.x;

  if ( dihedral_index < num_dihedrals ){
    unsigned dihedral_type = dtype[dihedral_index].x;

    unsigned a = dlist[dihedral_index].x;
    unsigned b = dlist[dihedral_index].y;
    unsigned c = dlist[dihedral_index].z;
    unsigned d = dlist[dihedral_index].w;

    unsigned offset = 6*dihedral_type;
    for ( int k=0; k<6; k++ ) p[k] = plist[offset+k];
  
    // Load the simulation box in local memory to avoid bank conflicts.
    float array[simulationBoxSize];
    simbox->loadMemory(array, simboxpointer);
  
    dr1 = simbox->calculateDistance(r[b], r[a], array);
    dr2 = simbox->calculateDistance(r[c], r[b], array);
    dr3 = simbox->calculateDistance(r[d], r[c], array);

    float c11 = dr1.w;
    float c12 = dr1.x*dr2.x + dr1.y*dr2.y + dr1.z*dr2.z;
    float c13 = dr1.x*dr3.x + dr1.y*dr3.y + dr1.z*dr3.z;
    float c22 = dr2.w;
    float c23 = dr2.x*dr3.x + dr2.y*dr3.y + dr2.z*dr3.z;
    float c33 = dr3.w;

    float cA = c13*c22 - c12*c23;
    float cB1 = c11*c22 - c12*c12;
    float cB2 = c22*c33 - c23*c23;
    float cD = sqrt(cB1*cB2);

    float cc = cA/cD;
  
    d_dihedrals[dihedral_index] = acos(cc);

    float epot = p[0] + (p[1]+(p[2]+(p[3]+(p[4]+p[5]*cc)*cc)*cc)*cc)*cc;

    if ( energy == 2 ){
      d_epot_dihedral[dihedral_index] = epot; 
    }
    else if ( energy == 1) {

      float t1 = cA;
      float t2 = c11*c23 - c12*c13;
      float t3 = -cB1;
      float t4 = cB2;
      float t5 = c13*c23 - c12*c33;
      float t6 = -cA;
      float cR1 = c12/c22; 
      float cR2 = c23/c22;

      float ff = -(p[1]+(2.f*p[2]+(3.f*p[3]+(4.f*p[4] + 5.f*p[5]*cc)*cc)*cc)*cc);
      
      ff *= c22;

      f1 = ff*(t1*dr1.x + t2*dr2.x + t3*dr3.x)/(cD*cB1);
      f2 = ff*(t4*dr1.x + t5*dr2.x + t6*dr3.x)/(cD*cB2);
    
      DihedralAtomicFloatAdd(&(f[a].x), f1);
      DihedralAtomicFloatAdd(&(f[b].x), -(1.f+cR1)*f1 + cR2*f2);
      DihedralAtomicFloatAdd(&(f[c].x), cR1*f1 - (1.f + cR2)*f2);
      DihedralAtomicFloatAdd(&(f[d].x), f2);    
    
      f1 = ff*(t1*dr1.y + t2*dr2.y + t3*dr3.y)/(cD*cB1);
      f2 = ff*(t4*dr1.y + t5*dr2.y + t6*dr3.y)/(cD*cB2);
    
      DihedralAtomicFloatAdd(&(f[a].y), f1);
      DihedralAtomicFloatAdd(&(f[b].y), -(1.f+cR1)*f1 + cR2*f2);
      DihedralAtomicFloatAdd(&(f[c].y), cR1*f1 - (1.f + cR2)*f2);
      DihedralAtomicFloatAdd(&(f[d].y), f2);    
    
      f1 = ff*(t1*dr1.z + t2*dr2.z + t3*dr3.z)/(cD*cB1);
      f2 = ff*(t4*dr1.z + t5*dr2.z + t6*dr3.z)/(cD*cB2);
 
      DihedralAtomicFloatAdd(&(f[a].z), f1);
      DihedralAtomicFloatAdd(&(f[b].z), -(1.f+cR1)*f1 + cR2*f2);
      DihedralAtomicFloatAdd(&(f[c].z), cR1*f1 - (1.f + cR2)*f2);
      DihedralAtomicFloatAdd(&(f[d].z), f2);    
    
      // Potential energy
      DihedralAtomicFloatAdd(&(f[a].w), epot);
 
    }
  }

}



/* From NVIDIA's programmers guide
__device__ void atomicFloatAdd(float* address, double val) { 

  float old = *address, assumed; 

  do { 
    assumed = old; 
    old = __int_as_float(atomicCAS((int*)address, __float_as_int(assumed), __float_as_int(val + assumed))); 
  } while (assumed != old); 

  return old; 
}
*/

