#include "rumd/MoleculeData.h"

#include <cstdlib>
#include <cmath>
#include <iostream>

#include "rumd/RUMD_Error.h"

///////////////////////////////////////////////////////
// Contructor / Destructor
///////////////////////////////////////////////////////

MoleculeData::MoleculeData(Sample *globalS, const char *fin){
  // The Sample class
  S = globalS;
 
  // The simulation box class
  SB = dynamic_cast<RectangularSimulationBox*>(S->GetSimulationBox());

  // Set flags
  flagHarmonic = 0;
  flagFENE = 0;
  flagConstraint = 0;
  flagAngle  = 0;
  flagDihedral  = 0;

  // Default max number of bonds
  max_num_bonds = 6;

  // Get number of molecules and max nuau from top file
  ReadTopology(fin);
 
  // Simply set the number of tpb to pbb (nice and cl comment!!)
  threads_per_block.x = S->GetParticlesPerBlock();
  num_blocks.x = S->GetNumberOfVirtualParticles()/threads_per_block.x;
 
  // Length of lists
  lblist = S->GetNumberOfVirtualParticles() + max_num_bonds*S->GetNumberOfParticles();
  lmlist = (max_num_uau + 1)*num_mol;
  lbtlist = max_num_bonds*S->GetNumberOfParticles();

  // blist: Allocate memory on host
  size_t nbytes = sizeof(uint1)*lblist;
  if ( cudaMallocHost((void **)&h_blist, nbytes) == cudaErrorMemoryAllocation )
    throw RUMD_Error("MoleculeData","MoleculeData","Memory allocation failure");
  
  // blist: Allocate memory on device
  if ( cudaMalloc((void **)&d_blist, nbytes) == cudaErrorMemoryAllocation)
    throw RUMD_Error("MoleculeData","MoleculeData","Memory allocation failure");
    
  // mlist: Allocate memory on host
  nbytes = sizeof(int1)*lmlist;
  if ( cudaMallocHost((void **)&h_mlist, nbytes) == cudaErrorMemoryAllocation)
    throw RUMD_Error("MoleculeData","MoleculeData","Memory allocation failure");
    
  // mlist: Allocate memory on the device
  if ( cudaMalloc((void **)&d_mlist, nbytes) == cudaErrorMemoryAllocation)
    throw RUMD_Error("MoleculeData","MoleculeData","Memory allocation failure");
 
  //btlist: On host   
  nbytes = sizeof(uint1)*lbtlist;
  if ( cudaMallocHost((void **)&h_btlist, nbytes) == cudaErrorMemoryAllocation)
    throw RUMD_Error("MoleculeData","MoleculeData","Memory allocation failure");

  // btlist: On device
  if ( cudaMalloc((void **)&d_btlist, nbytes) == cudaErrorMemoryAllocation)
    throw RUMD_Error("MoleculeData","MoleculeData","Memory allocation failure");
  
  // bplist: On host
  nbytes = sizeof(float2)*num_btypes;
  if ( cudaMallocHost((void **)&h_bplist, nbytes) == cudaErrorMemoryAllocation)
    throw RUMD_Error("MoleculeData","MoleculeData","Memory allocation failure");
  
  // bplist: On device
  if ( cudaMalloc((void **)&d_bplist, nbytes) == cudaErrorMemoryAllocation)
    throw RUMD_Error("MoleculeData","MoleculeData","Memory allocation failure");
 
  // btlist_int: On host
  nbytes = sizeof(int1)*num_btypes;
  if ( cudaMallocHost((void **)&h_btlist_int, nbytes) == cudaErrorMemoryAllocation)
    throw RUMD_Error("MoleculeData","MoleculeData","Memory allocation failure");
  
  // btlist_int: On device
  if ( cudaMalloc((void **)&d_btlist_int, nbytes) == cudaErrorMemoryAllocation)
    throw RUMD_Error("MoleculeData","MoleculeData","Memory allocation failure");
  

  // Data arrays 
  nbytes = sizeof(float4)*num_mol;
  if ( cudaMalloc((void **)&d_cm, nbytes) == cudaErrorMemoryAllocation || 
       cudaMallocHost((void **)&h_cm, nbytes) == cudaErrorMemoryAllocation )
    throw RUMD_Error("MoleculeData","MoleculeData","Memory allocation failure");

  nbytes = sizeof(float3)*num_mol;
  if ( cudaMalloc((void **)&d_vcm, nbytes) == cudaErrorMemoryAllocation || 
       cudaMallocHost((void **)&h_vcm, nbytes) == cudaErrorMemoryAllocation )
    throw RUMD_Error("MoleculeData","MoleculeData","Memory allocation failure");
  
  nbytes = sizeof(float3)*num_mol;
  if ( cudaMalloc((void **)&d_s, nbytes) == cudaErrorMemoryAllocation || 
       cudaMallocHost((void **)&h_s, nbytes) == cudaErrorMemoryAllocation )
    throw RUMD_Error("MoleculeData","MoleculeData","Memory allocation failure");
  
  nbytes = sizeof(float3)*num_mol;
  if ( cudaMalloc((void **)&d_omega, nbytes) == cudaErrorMemoryAllocation || 
       cudaMallocHost((void **)&h_omega, nbytes) == cudaErrorMemoryAllocation )
    throw RUMD_Error("MoleculeData","MoleculeData","Memory allocation failure");


  nbytes = sizeof(float)*num_mol*9;
  if ( cudaMalloc((void **)&d_inertia, nbytes) == cudaErrorMemoryAllocation || 
       cudaMallocHost((void **)&h_inertia, nbytes) == cudaErrorMemoryAllocation )
    throw RUMD_Error("MoleculeData","MoleculeData","Memory allocation failure");


  // Initialisation of blist (is important later on)
  for ( unsigned n=0; n<lblist; n++ ) h_blist[n].x = 0;
	
  // Initialisation of mlist (is important later on)
  for ( unsigned n=0; n<lmlist; n++ ) h_mlist[n].x = -1;

  // Initialisation of btlist_int (is important later on)
  for ( unsigned n=0; n<num_btypes; n++ ) h_btlist_int[n].x = -1;


  // Get mol. details from top file
  GetBondsFromTopology(fin); 
  GetAnglesFromTopology(fin);    // Memory is allocated here
  GetDihedralsFromTopology(fin); // Memory is allocated here

  // Calculate the molecular masses
  EvalMolMasses();
  
  // Print warning message about the off-diagonal components in
  // the stress tensor
  std::cout << "[Info]: RUMD does not support stress tensor calculations of molecular systems yet\n";
 
  
}


MoleculeData::~MoleculeData(){

  cudaFreeHost(h_mlist);
  cudaFree(d_mlist);

  cudaFreeHost(h_blist);
  cudaFree(d_blist);
  cudaFreeHost(h_btlist);
  cudaFree(d_btlist);
  cudaFreeHost(h_bplist);
  cudaFree(d_bplist);
  cudaFreeHost(h_btlist_int);
  cudaFree(d_btlist_int);
  cudaFreeHost(h_bonds);
  cudaFree(d_bonds);

  cudaFreeHost(h_alist);
  cudaFree(d_alist);
  cudaFreeHost(h_aplist);
  cudaFree(d_aplist);
  cudaFreeHost(h_epot_angle);
  cudaFree(d_epot_angle);
  cudaFreeHost(h_angles);
  cudaFree(d_angles);
 
  cudaFreeHost(h_dlist);
  cudaFree(d_dlist);
  cudaFreeHost(h_dtype);
  cudaFree(d_dtype);
  cudaFreeHost(h_epot_dihedral);
  cudaFree(d_epot_dihedral);
  cudaFreeHost(h_dihedrals);
  cudaFree(d_dihedrals);

  cudaFreeHost(h_cm);
  cudaFree(d_cm);
  cudaFreeHost(h_vcm);
  cudaFree(d_vcm);
  cudaFreeHost(h_s);
  cudaFree(d_s);
  cudaFreeHost(h_omega);
  cudaFree(d_omega);
  cudaFreeHost(h_inertia);
  cudaFree(d_inertia);

}

///////////////////////////////////////////////////////
// Methods.
///////////////////////////////////////////////////////

// Read the top file - get number of molecules and number of maximum united atomic units
void MoleculeData::ReadTopology(const char *str){
  char line[256];
  const char section[] = {'[', ' ', 'b', 'o', 'n', 'd', 's', ' ' , ']', '\n', '\0'};
  unsigned ai, aj, btype;
  int moli;
 
  // unsigned int num_part = S->GetNumberOfParticles();

  FILE *top = fopen(str, "r");
  if ( top == NULL )
    throw RUMD_Error("MoleculeData","GetBondsFromTopology",
		     "Couldn't open topology file.");

  // Set the file pointer to the line after 
  // the [ bond ] section string
  do { 
    fgets(line, 256, top);
  } while ( strcmp(line, section) != 0 ); 
  
  // Skip next line (comments)
  fgets(line, 256, top);

  int nmol = 0; num_btypes=0;
  int *marray = (int*)malloc(0);

  int ii=0;
  // Go through the pair list
  do {
    
    // Check that we have an entry
    fpos_t fpos;
    fgetpos(top, &fpos);    
    fgets(line, 256, top);

    // New section: break out
    if ( line[0] == '[' )  break;

    // If so, reset pointer to line start and read again
    // but with format
    fsetpos(top, &fpos);
    if ( fscanf(top, "%u%u%u%u\n", &moli, &ai, &aj, &btype) != 4 ) {
      throw RUMD_Error("MoleculeData","GetBondsFromTopology",
		       "Format not correct");
    }

    if ( btype > num_btypes ) num_btypes++;
 
    if ( moli>nmol-1 ){
      nmol = moli+1;
      marray = (int*)realloc(marray, nmol*sizeof(int));
      if ( marray == NULL )
	 throw RUMD_Error("MoleculeData","GetBondsFromTopology", 
			  "Couldn't allocate memory");
      marray[moli] = 0;
    }
    marray[moli]++; 
    ii++;
  } while ( !feof(top) );
  
  fclose(top);

  int maxnuau = 0; 
  for ( int n=0; n<nmol; n++ ){ //nmol
    if ( marray[n] > maxnuau ) maxnuau = marray[n];
  }
  
  maxnuau++;
  num_btypes ++;

  num_mol = nmol;
  max_num_uau = maxnuau; 

  std::cout << "Topology file "<< str << " found. \n";  
  std::cout << "Read "<< num_mol << " molecules with " 
	    << max_num_uau << " maximum united atomic units. " ;
  std::cout << "Read " << num_btypes << " bonding type(s)."<< std::endl;


  // Allocate for storing the actual/computed bond lengths
  unsigned nbytes = sizeof(float)*ii;
  if ( ( h_bonds = (float *)malloc(nbytes) ) == NULL ||  
       cudaMalloc((void **)&d_bonds, nbytes) == cudaErrorMemoryAllocation )
    throw RUMD_Error("MoleculeData",__func__, "Memory allocation failure");
  
  free(marray);
}


// Set molecule lists from top file
void MoleculeData::GetBondsFromTopology(const char *str){
  char line[256];
  const char section[] = {'[', ' ', 'b', 'o', 'n', 'd', 's', ' ' , ']', '\n', '\0'};
  unsigned ai, aj, moli, btype;
  
  unsigned *mlist_index = new unsigned [num_mol];
  for ( unsigned n=0; n<num_mol; n++ ) mlist_index[n] = 0;
  
  unsigned *btlist_index = new unsigned [lbtlist];
  for ( unsigned n=0; n<lbtlist; n++ ) btlist_index[n] = 0;
  
  FILE *top = fopen(str, "r");
  if ( top == NULL )
    throw RUMD_Error("MoleculeData",__func__,"Couldn't open topology file.");

  // Set the file pointer to the line after 
  // the [ bond ] section string
  do { 
    fgets(line, 256, top);
  } while ( strcmp(line, section) != 0 ); 
  
  // Skip next line (comments)
  fgets(line, 256, top);

  // Go through the pair list
  do {
    
    // Check that we have an entry
    fpos_t fpos;
    fgetpos(top, &fpos);    
    fgets(line, 256, top);
    if ( line[0] == '[' ) break; 

    // If so reset pointer to line start and read again
    // but with format
    fsetpos(top, &fpos);
    if ( fscanf(top, "%u%u%u%u\n", &moli, &ai, &aj, &btype) != 4 ) 
      throw RUMD_Error("MoleculeData",__func__,
			"Incorrect file format");

    CheckRead(moli, ai, aj);

    // Add entries to blist
    AddblistEntry(ai, aj);

    // Add entries to mlist
    AddmlistEntry(mlist_index, moli, ai, aj);

    // Add entries into btlist 
    AddbtlistEntry(btlist_index, ai, aj, btype);

  } while ( !feof(top) );
  
  fclose(top);

  
  // Copy the arrays on to the device (Checks are missing)
  size_t nbytes = sizeof(uint1)*lblist;
  cudaMemcpy(d_blist, h_blist, nbytes, cudaMemcpyHostToDevice);
  
  nbytes = sizeof(int1)*lmlist;
  cudaMemcpy(d_mlist, h_mlist, nbytes, cudaMemcpyHostToDevice);

  nbytes = sizeof(int1)*lbtlist;
  cudaMemcpy(d_btlist, h_btlist, nbytes, cudaMemcpyHostToDevice);

  // Freeing the mlist index array
  delete[] mlist_index;
  delete[] btlist_index;

}

// Helper function for top-reader
void MoleculeData::CheckRead(unsigned i, unsigned ai, unsigned aj){
  if ( i >= num_mol )
    throw RUMD_Error("MoleculeData","CheckRead", "Molecule index in .top file is larger than number of molecules.");
  // MODIFY THIS.
  if ( ai >= S->GetNumberOfParticles() || aj >= S->GetNumberOfParticles() ) 
    throw RUMD_Error("MoleculeData","CheckRead", "Particle index in .top file is larger than number of particles\n.");
}

// Read the top file - get number of angles: IMPORTANT must be called after the
// GetNumBondsFromTopology function
void MoleculeData::GetAnglesFromTopology(const char *str){
  char line[256];
  const char section[] = {'[', ' ', 'a', 'n', 'g', 'l', 'e', 's' , ' ', ']', '\n', '\0'};
  unsigned a, b, c, atype, moli;
 
  unsigned num_part = S->GetNumberOfParticles();
  num_atypes = 0;
  num_angles = 0;

  h_alist = (uint4 *)malloc(num_angles*sizeof(uint4)); 
  if ( h_alist == NULL )
    throw RUMD_Error("MoleculeData","GetAnglesFromTopology",
		     "Couldn't allocate memory.");


  FILE *top = fopen(str, "r");
  if ( top == NULL )
    throw RUMD_Error("MoleculeData","GetAnglesFromTopology",
		     "Couldn't open topology file.");

  // Set the file pointer to the line after 
  // the [ angles ] section string
  do { 
    fgets(line, 256, top);

    if ( feof(top) ){
      fclose(top);
      return;
    }

  } while ( strcmp(line, section) != 0 ); 
  

  // Skip next line (comments)
  fgets(line, 256, top);

  // Go through the pair list
  do {
    
    // Check that we have an entry
    fpos_t fpos;
    fgetpos(top, &fpos);    
    fgets(line, 256, top);
    if ( line[0] == '[' ) break; 

    // If so, reset pointer to line start and read again
    // but with format
    fsetpos(top, &fpos);
    if ( fscanf(top, "%u%u%u%u%u\n", &moli, &a, &b, &c, &atype) !=5 )
      throw RUMD_Error("MoleculeData","GetAnglesFromTopology",
		       "Format not correct");
    
    if ( moli>num_mol-1 || a>num_part-1 || 
	 b>num_part-1 || c>num_part-1 )
      throw RUMD_Error("MoleculeData","GetAnglesFromTopology",
		       "Indicies are too large");
 
    num_angles ++;

    // Reallocate memory
    h_alist = (uint4 *)realloc(h_alist, sizeof(uint4)*num_angles);
    if ( h_alist == NULL )
      throw RUMD_Error("MoleculeData","GetAnglesFromTopology",
		       "Couldn't allocate memory.");

    h_alist[num_angles-1].x = a;
    h_alist[num_angles-1].y = b;
    h_alist[num_angles-1].z = c;
    h_alist[num_angles-1].w = atype;
    
    if ( atype > num_atypes ) num_atypes++;
     
  } while ( !feof(top) );
  
  fclose(top);

  num_atypes ++;
  
  // Allocating device memory
  size_t nbytes = sizeof(uint4)*num_angles;
  if ( cudaMalloc((void **)&d_alist, nbytes) == cudaErrorMemoryAllocation )
    throw RUMD_Error("MoleculeData","GetAnglesFromTopology","Memory allocation failure");

  // Copy angle info
  cudaMemcpy(d_alist, h_alist, nbytes, cudaMemcpyHostToDevice);

  // Allocate memory for parameters
  nbytes = sizeof(float2)*num_atypes;
  if ( ( h_aplist = (float2 *)malloc(nbytes) ) == NULL ||  
       cudaMalloc((void **)&d_aplist, nbytes) == cudaErrorMemoryAllocation )
    throw RUMD_Error("MoleculeData","GetAnglesFromTopology","Memory allocation failure");
  
  
  // Allocate for potential energy arrays
  nbytes = sizeof(float)*num_angles;
  if ( ( h_epot_angle = (float *)malloc(nbytes) ) == NULL ||  
       cudaMalloc((void **)&d_epot_angle, nbytes) == cudaErrorMemoryAllocation )
    throw RUMD_Error("MoleculeData","GetAnglesFromTopology","Memory allocation failure");

  
  // Allocate for storing the actual/computed angles
  if ( ( h_angles = (float *)malloc(nbytes) ) == NULL ||  
       cudaMalloc((void **)&d_angles, nbytes) == cudaErrorMemoryAllocation )
    throw RUMD_Error("MoleculeData","GetAnglesFromTopology","Memory allocation failure");
  

  std::cout << "Read "<< num_angles << " angle(s). " ;
  std::cout << "Read " << num_atypes << " angle type(s)."<< std::endl;

  
}


// Read the top file - get number of angles: IMPORTANT must be called after the
// GetNumBondsFromTopology function
void MoleculeData::GetDihedralsFromTopology(const char *str){
  char line[256];
  const char section[] = {'[', ' ', 'd', 'i', 'h', 'e', 'd', 'r' , 'a', 'l', 's', ' ', ']', '\n', '\0'};
  unsigned a, b, c, d, dtype, moli;
 
  unsigned num_part = S->GetNumberOfParticles();
  num_dtypes = 0;
  num_dihedrals = 0;

  h_dlist = (uint4 *)malloc(num_angles*sizeof(uint4)); 
  h_dtype = (uint1 *)malloc(num_angles*sizeof(uint1));
  if ( h_dlist == NULL || h_dtype == NULL )
    throw RUMD_Error("MoleculeData","GetDihedralsFromTopology",
		     "Couldn't allocate memory.");


  FILE *top = fopen(str, "r");
  if ( top == NULL )
    throw RUMD_Error("MoleculeData","GetDihedralsFromTopology",
		     "Couldn't open topology file.");

  // Set the file pointer to the line after 
  // the [ dihedrals ] section string
  do { 
    fgets(line, 256, top);

    if ( feof(top) ){
      fclose(top);
      return;
    }

  } while ( strcmp(line, section) != 0 ); 
  

  // Skip next line (comments)
  fgets(line, 256, top);

  // Go through the pair list
  do {
    
    // Check that we have an entry
    fpos_t fpos;
    fgetpos(top, &fpos);    
    fgets(line, 256, top);
    if ( line[0] == '[' ) break; 

    // If so, reset pointer to line start and read again
    // but with format
    fsetpos(top, &fpos);
    if ( fscanf(top, "%u%u%u%u%u%u\n", &moli, &a, &b, &c, &d, &dtype) != 6 )
      throw RUMD_Error("MoleculeData","GetDihedralsFromTopology",
		       "Format not correct");
    
    if ( moli>num_mol-1 || a>num_part-1 || 
	 b>num_part-1 || c>num_part-1 || d>num_part-1)
      throw RUMD_Error("MoleculeData","GetDihedralsFromTopology",
		       "Indicies are too large");
 
    num_dihedrals ++;

    // Reallocate memory
    h_dlist = (uint4 *)realloc(h_dlist, sizeof(uint4)*num_dihedrals);
    h_dtype = (uint1 *)realloc(h_dtype, sizeof(uint1)*num_dihedrals);
    if ( h_dlist == NULL || h_dtype == NULL )
      throw RUMD_Error("MoleculeData","GetDihedralsFromTopology",
		       "Couldn't allocate memory.");

    h_dlist[num_dihedrals-1].x = a;
    h_dlist[num_dihedrals-1].y = b;
    h_dlist[num_dihedrals-1].z = c;
    h_dlist[num_dihedrals-1].w = d;

    h_dtype[num_dihedrals-1].x = dtype;
    
    if ( dtype > num_dtypes ) num_dtypes++;
     
  } while ( !feof(top) );
  
  fclose(top);

  num_dtypes ++;
  
  // Copy the info to device
  size_t nbytes = sizeof(uint4)*num_dihedrals;
  if ( cudaMalloc((void **)&d_dlist, nbytes) == cudaErrorMemoryAllocation )
    throw RUMD_Error("MoleculeData","GetDihedralsFromTopology","Memory allocation failure");

  cudaMemcpy(d_dlist, h_dlist, nbytes, cudaMemcpyHostToDevice);


  nbytes = sizeof(uint1)*num_dihedrals;
  if ( cudaMalloc((void **)&d_dtype, nbytes) == cudaErrorMemoryAllocation )
    throw RUMD_Error("MoleculeData","GetDihedralsFromTopology","Memory allocation failure");

  cudaMemcpy(d_dtype, h_dtype, nbytes, cudaMemcpyHostToDevice);
    
  
  // Allocate memory for parameter lists (both host and device) - 6 because 
  // of the Ryckaert-Bellemans potential
  nbytes = sizeof(float)*6*num_dtypes;
  if ( ( h_dplist = (float *)malloc(nbytes) ) == NULL ||  
       cudaMalloc((void **)&d_dplist, nbytes) == cudaErrorMemoryAllocation )
    throw RUMD_Error("MoleculeData","GetDihedralsFromTopology","Memory allocation failure");
  
  
  // Allocate for potential energy arrays
  nbytes = sizeof(float)*num_dihedrals;
  if ( ( h_epot_dihedral = (float *)malloc(nbytes) ) == NULL ||  
       cudaMalloc((void **)&d_epot_dihedral, nbytes) == cudaErrorMemoryAllocation )
    throw RUMD_Error("MoleculeData","GetDihedralsFromTopology","Memory allocation failure");

  // Allocate for storing the actual/computed dihedrals
  if ( ( h_dihedrals = (float *)malloc(nbytes) ) == NULL ||  
       cudaMalloc((void **)&d_dihedrals, nbytes) == cudaErrorMemoryAllocation )
    throw RUMD_Error("MoleculeData","GetDihedralsFromTopology","Memory allocation failure");
  

  std::cout << "Read "<< num_dihedrals<< " dihedrals(s). " ;
  std::cout << "Read " << num_dtypes << " dihedral type(s)."<< std::endl;

  
}

///////////////////////////////////////////////////////
// Add from topology
///////////////////////////////////////////////////////

void MoleculeData::AddbtlistEntry(unsigned *index, unsigned ai, unsigned aj, unsigned type){
  if ( index[ai]==max_num_bonds || index[aj]==max_num_bonds )
      throw RUMD_Error("MoleculeData","ReadTotology", "Number of bonds has exceeded the max_num_bonds");

  h_btlist[ai*max_num_bonds + index[ai]].x = type;
  h_btlist[aj*max_num_bonds + index[aj]].x = type;
  
  index[ai] = index[ai] + 1;
  index[aj] = index[aj] + 1;
}

// Helper function for top-reader
void MoleculeData::AddblistEntry(unsigned ai, unsigned aj){
  unsigned i = S->GetNumberOfVirtualParticles() + ai*max_num_bonds + h_blist[ai].x;  
  h_blist[i].x = aj;
  
  i = S->GetNumberOfVirtualParticles() + aj*max_num_bonds + h_blist[aj].x;  
  h_blist[i].x = ai;
  
  h_blist[ai].x++;
  h_blist[aj].x++;
}

// Helper function for top-reader
void MoleculeData::AddmlistEntry(unsigned *index, unsigned i, 
				 unsigned ai, unsigned aj){
  int flag[2] = {0};
  unsigned si = i*max_num_uau;
  
  for ( unsigned n=0; n<index[i]; n++ ){
    if ( h_mlist[si+n+num_mol].x == (signed)ai ) flag[0] = 1;
    if ( h_mlist[si+n+num_mol].x == (signed)aj ) flag[1] = 1;
  }

  if ( flag[0] == 0 ){
    h_mlist[si+index[i]+num_mol].x = ai;
    index[i]++;
  }
  
  if ( flag[1] == 0 ){
    h_mlist[si+index[i]+num_mol].x = aj;
    index[i]++;
  }
   
  h_mlist[i].x = (signed)index[i];
}

///////////////////////////////////////////////////////
// Set methods
///////////////////////////////////////////////////////

void MoleculeData::SetBondHarmonic( unsigned type, float lbond, float ks ){
  if( flagHarmonic == 0 ){
    HarmBond = new BondHarmonic(this, this->S);
    S->AddPotential(HarmBond);
    flagHarmonic = 1;
  }
  
  if ( type>num_btypes )
    throw RUMD_Error("MoleculeData","SetBondHarmonic", "Type index too large\n.");
  
  h_bplist[type].x = lbond;
  h_bplist[type].y = ks;
  
  h_btlist_int[type].x = 0;
  
  size_t nbytes = sizeof(float2)*num_btypes;
  cudaMemcpy(d_bplist, h_bplist, nbytes, cudaMemcpyHostToDevice);
  
  nbytes = sizeof(int1)*num_btypes;
  cudaMemcpy(d_btlist_int, h_btlist_int, nbytes, cudaMemcpyHostToDevice);

}

void MoleculeData::SetBondConstraint( unsigned bondType, float bondLength ){
  if( bondType > num_btypes || bondLength <= 0 )
    throw RUMD_Error("MoleculeData", "SetBondConstraint", "Type index too large or bond-length is negative.\n");
  
  // Update internal host and device side data.
  h_btlist_int[bondType].x = 2; 
  h_bplist[bondType].x = bondLength; 
  h_bplist[bondType].y = 0; // Not used.
  
  cudaMemcpy( d_btlist_int, h_btlist_int, num_btypes * sizeof(int1), cudaMemcpyHostToDevice );  
  cudaMemcpy( d_bplist, h_bplist, num_btypes * sizeof(float2), cudaMemcpyHostToDevice );
  
  unsigned maxBonds = GetMaximumNumberOfBonds();
  
  // Count the number of constraints;
  unsigned count = 0;
  for( unsigned i = 0; i < S->GetNumberOfParticles(); i++){
    unsigned realBonds = h_blist[i].x;
    
    for( unsigned j = 0; j < realBonds; j++){
      unsigned userBondType = h_btlist[i*maxBonds+j].x;
      unsigned internalBondType = h_btlist_int[userBondType].x;
      
      if( userBondType == bondType && internalBondType == 2 )
	count++;
    }
  }
  
  // Subtract 1 degree of freedom for each constraint.
  S->SubtractDOFs( count / 2 );
  
  if( flagConstraint == 0 ){
    constraintPotential = new ConstraintPotential(this, this->S);
    S->AddPotential(constraintPotential);
    flagConstraint = 1;
  }
  
  constraintPotential->UpdateConstraintLayout();
}

void MoleculeData::SetBondFENE(unsigned type){
  if( flagFENE == 0 ){
    FENEBond = new BondFENE(this, this->S);
    S->AddPotential(FENEBond);
    flagFENE = 1;
  }
  
  if ( type>num_btypes )
    throw RUMD_Error("MoleculeData","SetBondHarmonic", "Type index too large\n.");

  h_btlist_int[type].x = 1;

  size_t nbytes = sizeof(int1)*num_btypes;
  cudaMemcpy(d_btlist_int, h_btlist_int, nbytes, cudaMemcpyHostToDevice);
}


void MoleculeData::SetAngleCosSq( unsigned type, float mtheta0, float mktheta ){

  if ( num_angles == 0 )
    throw RUMD_Error("MoleculeData","SetAngleCosSq", "No angles defined in top-file\n.");

  if( flagAngle == 0 ){
    AngleCos = new AngleCosSq(this, this->S);
    S->AddPotential(AngleCos);
    flagAngle = 1;
  }

  h_aplist[type].x = mtheta0;
  h_aplist[type].y = mktheta;

  size_t nbytes = sizeof(float2)*num_atypes;
  cudaMemcpy(d_aplist, h_aplist, nbytes, cudaMemcpyHostToDevice);

  
}

void MoleculeData::SetDihedralRyckaert( unsigned type, float c0, 
					float c1, float c2, float c3, 
					float c4, float c5 ){

  if ( num_dihedrals == 0 )
    throw RUMD_Error("MoleculeData","SetDihedralRyckaert", 
		     "No dihedrals defined in top-file\n.");


  if( flagDihedral == 0 ){
    DihedralRyc = new DihedralRyckaert(this, this->S);
    S->AddPotential(DihedralRyc);
    flagDihedral = 1;
  }
  
  unsigned offset = type*6;

  h_dplist[offset]   = c0;
  h_dplist[offset+1] = c1;
  h_dplist[offset+2] = c2;
  h_dplist[offset+3] = c3;
  h_dplist[offset+4] = c4;
  h_dplist[offset+5] = c5;
    
  size_t nbytes = sizeof(float)*6*num_dtypes;
  cudaMemcpy(d_dplist, h_dplist, nbytes, cudaMemcpyHostToDevice);
  
}


void MoleculeData::SetExclusionBond(PairPotential* potential, unsigned etype){

  potential->SetExclusionBond(h_btlist, h_blist, S->GetNumberOfParticles(), 
			      S->GetNumberOfVirtualParticles(), max_num_bonds, etype);

}

void MoleculeData::SetExclusionMolecule(PairPotential *potential, unsigned molindex){
  potential->SetExclusionMolecule(h_mlist, molindex, max_num_uau);
}

void MoleculeData::SetExclusionMolecule(PairPotential *potential){
  for ( unsigned n=0; n<num_mol; n++ ) potential->SetExclusionMolecule(h_mlist, n, max_num_uau);
}

void MoleculeData::SetExclusionAngle(PairPotential* potential){

  potential->SetExclusionAngle(h_alist, num_angles);

}

void MoleculeData::SetExclusionDihedral(PairPotential* potential){

  potential->SetExclusionDihedral(h_dlist, num_dihedrals);

}

void MoleculeData::SetHarmonicBondConstant(unsigned type, float ks){
  
  h_bplist[type].y = ks;
 
  size_t nbytes = sizeof(float2)*num_btypes;
  cudaMemcpy(d_bplist, h_bplist, nbytes, cudaMemcpyHostToDevice);
 

}

void MoleculeData::SetAngleConstant(unsigned type, float ktheta){

  h_aplist[type].y = ktheta;

  size_t nbytes = sizeof(float2)*num_atypes;
  cudaMemcpy(d_aplist, h_aplist, nbytes, cudaMemcpyHostToDevice);

}

void MoleculeData::SetDihedralRyckaertConstants(unsigned type, float c0, float c1, float c2, 
						float c3, float c4, float c5){

  
  unsigned offset = type*6;

  h_dplist[offset]   = c0;
  h_dplist[offset+1] = c1;
  h_dplist[offset+2] = c2;
  h_dplist[offset+3] = c3;
  h_dplist[offset+4] = c4;
  h_dplist[offset+5] = c5;
    
  size_t nbytes = sizeof(float)*6*num_dtypes;
  cudaMemcpy(d_dplist, h_dplist, nbytes, cudaMemcpyHostToDevice);
  
}


///////////////////////////////////////////////////////
// Debuggin methods
///////////////////////////////////////////////////////

// Print mlist and blist (for debugging mostly)
void MoleculeData::PrintLists(void){

  std::cout << "--- blist: Sec I ---" << std::endl;
  for ( unsigned i=0; i<S->GetNumberOfVirtualParticles(); i++ )
    std::cout << h_blist[i].x << std::endl;
  
  std::cout << "--- blist: Sec II ---" << std::endl;
  for ( unsigned i=S->GetNumberOfVirtualParticles(); i<lblist; i++ )
    std::cout << h_blist[i].x << std::endl;
	
  std::cout << "--- mlist: ---" << std::endl;
  for ( unsigned i=0; i<lmlist; i++)
    std::cout << h_mlist[i].x << std::endl;
}


// Print mlist and blist for i (for debugging mostly)
void MoleculeData::PrintLists(unsigned i){
  
  if(i>S->GetNumberOfParticles())
    std::cout << "index " << i << " too large (number of particles is " << S->GetNumberOfParticles() << ")" << std::endl; 
    else {
  std::cout << "Atom index " << i << " has " <<  h_blist[i].x 
	    << " bonds." << " Binding indices "; 
  for ( unsigned n=0; n<h_blist[i].x; n++ )
    std::cout << h_blist[S->GetNumberOfVirtualParticles() 
			 + i*max_num_bonds + n].x << " ";
  std::cout << std::endl;
  }
}

// Print mlist to file - convenient for post run data analysis
void MoleculeData::PrintMlist(){
  FILE *fout = fopen("mlist.top", "w");

  if ( fout == NULL )
    throw RUMD_Error("MoleculeData","PrintMlist",
		     "Couldn't open file for writing - something seriously wrong!");

  std::cout << "Writing mlist to mlist.top\n";

  for ( unsigned n=0; n<lmlist; n++ )
    fprintf(fout, "%d\n", h_mlist[n].x);

  fclose(fout);
}



/////////////////////////////////
// Eval methods
////////////////////////////////////

void MoleculeData::EvalMolMasses(void){

  S->particleData.CopyVelFromDevice();
    
  for ( unsigned n=0; n<num_mol; n++ )  h_cm[n].w = 0.0f;
  
  for ( unsigned n=0; n<num_mol; n++ ){
    unsigned offset = max_num_uau*n + num_mol;
    
    for ( int m=0; m<h_mlist[n].x; m++ ){
      unsigned a = h_mlist[offset+m].x;
      h_cm[n].w += 1.0/S->particleData.h_v[a].w;
    }
  }

}


// Could be implemented as device function
void MoleculeData::EvalVelCM(void){

  S->particleData.CopyVelFromDevice();
 
  for ( unsigned n=0; n<num_mol; n++ ) 
    h_vcm[n].x = h_vcm[n].y = h_vcm[n].z = 0.0f;

  for ( unsigned n=0; n<num_mol; n++ ){

    unsigned offset = max_num_uau*n + num_mol;
    for ( int m=0; m<h_mlist[n].x; m++ ){
      unsigned a = h_mlist[offset+m].x;
      float mass = 1.0/S->particleData.h_v[a].w;

      h_vcm[n].x += S->particleData.h_v[a].x*mass;
      h_vcm[n].y += S->particleData.h_v[a].y*mass;
      h_vcm[n].z += S->particleData.h_v[a].z*mass;
    }

    h_vcm[n].x /= h_cm[n].w;
    h_vcm[n].y /= h_cm[n].w;
    h_vcm[n].z /= h_cm[n].w;
  }

} 



// Could be implemented as device function
void MoleculeData::EvalCM(void){

  float4 r={0.0f, 0.0f, 0.0f, 0.0f};

  for ( unsigned n=0; n<num_mol; n++ ) 
    h_cm[n].x = h_cm[n].y = h_cm[n].z = 0.0f;
  

  S->particleData.CopyPosFromDevice();
  
  for ( unsigned n=0; n<num_mol; n++ ){

    unsigned offset = max_num_uau*n + num_mol;
    unsigned nuau   = h_mlist[n].x;

    unsigned i = h_mlist[offset].x;
    float x = S->particleData.h_r[i].x;
    float y = S->particleData.h_r[i].y;
    float z = S->particleData.h_r[i].z;

    float mi = 1.0/S->particleData.h_v[i].w;      

    h_cm[n].x += mi*x;
    h_cm[n].y += mi*y;
    h_cm[n].z += mi*z;

    for ( unsigned m=1; m<nuau; m++ ){
      i = h_mlist[offset+m].x;
      unsigned j = h_mlist[offset+m-1].x;
      
      r = SB->calculateDistance(S->particleData.h_r[i], 
				S->particleData.h_r[j], 
				SB->GetHostPointer());

      x += r.x; y += r.y; z += r.z;

      mi = 1.0/S->particleData.h_v[i].w;      
      h_cm[n].x += mi*x; h_cm[n].y += mi*y;  h_cm[n].z += mi*z;

    }

    h_cm[n].x /= h_cm[n].w;
    h_cm[n].y /= h_cm[n].w;
    h_cm[n].z /= h_cm[n].w;
  
  }

}


// Could-Should be implemented as device function
void MoleculeData::EvalSpin(void){


  S->particleData.CopyVelFromDevice();
  EvalCM();

  for ( unsigned n=0; n<num_mol; n++ ){
    
    unsigned nuau   = h_mlist[n].x;
    unsigned offset = max_num_uau*n + num_mol;
    unsigned i_offset = n*9;
    
    h_s[n].x = 0.0; h_s[n].y = 0.0; h_s[n].z = 0.0;
    for ( int k=0; k<9; k++ ) h_inertia[i_offset + k] = 0.0;
    
    for ( unsigned i=0; i<nuau; i++ ){

      unsigned in = h_mlist[offset+i].x;
      float mass = 1.0/S->particleData.h_v[in].w;
   
      float4 dcm = SB->calculateDistance(S->particleData.h_r[in], h_cm[n], 
					 SB->GetHostPointer());

      float4 vel = S->particleData.h_v[in];
      
      // Angular momentum
      h_s[n].x += dcm.y*mass*vel.z - dcm.z*mass*vel.y;
      h_s[n].y += dcm.z*mass*vel.x - dcm.x*mass*vel.z;
      h_s[n].z += dcm.x*mass*vel.y - dcm.y*mass*vel.x;

      // Inertia tensor
      float dcmx2 = mass*dcm.x*dcm.x;  
      float dcmy2 = mass*dcm.y*dcm.y; 
      float dcmz2 = mass*dcm.z*dcm.z;

      h_inertia[i_offset]   += dcmy2 + dcmz2;
      h_inertia[i_offset+4] += dcmx2 + dcmz2;
      h_inertia[i_offset+8] += dcmy2 + dcmx2;

      h_inertia[i_offset+1] -= mass*dcm.x*dcm.y;
      h_inertia[i_offset+2] -= mass*dcm.x*dcm.z;
      h_inertia[i_offset+5] -= mass*dcm.y*dcm.z;
    }

    // Symmetri 
    h_inertia[i_offset+3] = h_inertia[i_offset+1];
    h_inertia[i_offset+6] = h_inertia[i_offset+2];
    h_inertia[i_offset+7] = h_inertia[i_offset+5];

    // Matrix rank should be checked: det neq 0 -
    if ( fabs(Determinant3(&h_inertia[i_offset])) < FLT_EPSILON ){
      float inert =  h_inertia[i_offset] + h_inertia[i_offset+4] +  h_inertia[i_offset+8];
      inert /= 3.0;

      h_omega[n].x = h_s[n].x/inert;
      h_omega[n].y = h_s[n].y/inert;
      h_omega[n].z = h_s[n].z/inert;
    }
    else {
      float s[3], omega[3];  // Solve for ang. velocity

      s[0] = h_s[n].x; s[1] = h_s[n].y; s[2] = h_s[n].z;

      GaussPivotBackCPU(omega, s, &h_inertia[i_offset], 3);
      
      h_omega[n].x = omega[0]; h_omega[n].y = omega[1]; h_omega[n].z = omega[2];
            
    }

  }

}



///////////////////////////////////////
// Contraint stuff
//////////////////////////////////////

void MoleculeData::SetNumberOfConstraintIterations( unsigned numIte ){ 

  constraintPotential->SetNumberOfConstraintIterations(numIte); 

}

void MoleculeData::WriteConstraints( void ){ 

  constraintPotential->WritePotential(S->GetSimulationBox()); 

}

///////////////////////////////////
// Write and Get methods
///////////////////////////////

void MoleculeData::WriteMolConf(const char *fstr){

  EvalVelCM();
  EvalCM();

  float4 box = SB->GetSimulationBox();

  FILE *fptr = fopen(fstr, "w");
  if ( fptr == NULL )
    throw RUMD_Error("MoleculeData", __func__, "Couldn't open file\n.");

  fprintf(fptr, "%d\n", num_mol);
  fprintf(fptr, "ioformat=1 boxLengths=%f,%f,%f columns=type,x,y,z,vx,vy,vz\n", 
	  box.x, box.y, box.z);

  // Note: Molecule types not implemented yet.
  for ( unsigned n=0; n<num_mol; n++ ){
    fprintf(fptr, "0 %f %f %f %f %f %f\n", h_cm[n].x, h_cm[n].y, h_cm[n].z, 
	    h_vcm[n].x, h_vcm[n].y, h_vcm[n].z);
  } 
 
  fclose(fptr);
}




void MoleculeData::GetAngles( void ){
  
  size_t nbytes = num_angles*sizeof(float);
  cudaMemcpy(h_angles, d_angles, nbytes, cudaMemcpyDeviceToHost);
  
}


void MoleculeData::GetDihedrals( void ){
  
  size_t nbytes = num_dihedrals*sizeof(float);
  cudaMemcpy(h_dihedrals, d_dihedrals, nbytes, cudaMemcpyDeviceToHost);

}

void MoleculeData::WriteMolxyz(const char *fstr, unsigned mol_id, 
			       const char mode[2]){
 
    
  FILE *fout = fopen(fstr, mode);
  if ( fout == NULL )
    throw RUMD_Error("MoleculeData", __func__, "Couldn't open file\n.");

  
  S->particleData.CopyPosFromDevice();
  
  unsigned nuau = h_mlist[mol_id].x;
  
  fprintf(fout, "%u\n", nuau);
  fprintf(fout, "\n");

  size_t offset = max_num_uau*mol_id + num_mol;
  for ( unsigned i=offset; i<offset+nuau; i++ ){
    unsigned j = h_mlist[i].x;
    fprintf(fout, "%d %f %f %f\n", S->particleData.h_Type[j], 
	    S->particleData.h_r[j].x, S->particleData.h_r[j].y, 
	    S->particleData.h_r[j].z);
  }

  fclose(fout);

}

/////////////////////////////////////////////////
/// Helpers - should be placed in a generic place
////////////////////////////////////////////////


void GaussPivotBackCPU(float *x, float *b, float *A, unsigned nrc){
  unsigned i=0, j=0, k, kk; 
  float tmp, c;

  for ( unsigned int n=0; n<nrc; n++ ) x[n] = b[n];

  while ( i<nrc && j<nrc ){
    
    // Find pivot
    unsigned ir = i;
    for ( k=i+1; k<nrc; k++ ){
      if ( fabs(A[k*nrc+j]) > fabs(A[ir*nrc+j]) )
	ir = k;
    }
    
    if ( fabs(A[ir*nrc+j])>FLT_EPSILON ){

      // Swap rows
      for ( k=0; k<nrc; k++ ){
	tmp = A[i*nrc+k];
	A[i*nrc+k] = A[ir*nrc+k];
	A[ir*nrc+k] = tmp;
      }
      
      tmp = b[i];
      b[i] = b[ir];
      b[ir] = tmp;

      // Divide row i (Eq. (i)) with A(i,j)
      c = A[i*nrc+j];
      b[i] /= c;
      for ( k=0; k<nrc; k++ ) A[i*nrc+k] /= c;
            
      // Multiply and substract A(i,j) - Gaussian part
      for ( k=i+1; k<nrc; k++ ){
	c = A[k*nrc + j];
	for ( kk=0; kk<3; kk++ )
	  A[k*nrc+kk] -= c*A[i*nrc+kk];
	b[k] -= c*b[i];
      }

      i++;
    }

    j++;
  }

  // Backsubstitution
  for ( int i=nrc-1; i>=0; i-- ){
    
    float s = 0.0;
    for ( int j=i+1; j<(signed)nrc; j++ )
      s += A[i*nrc+j]*b[j];

    b[i] = (b[i] - s)/A[i*nrc+i];
  }

  
  for ( unsigned int n=0; n<nrc; n++ ){
    float tmp = x[n];
    x[n] = b[n];
    b[n] = tmp;
  }

}

double Determinant3(float *A){
  

  double a = A[0]*A[4]*A[8];
  double b = A[0]*A[5]*A[7];
  double c = A[1]*A[3]*A[8];
  double d = A[1]*A[4]*A[6];
  double e = A[2]*A[3]*A[7];
  double f = A[2]*A[4]*A[6];

  return a - b - c + d + e - f;

}

