
/*
  Copyright (C) 2010  Thomas Schrøder
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  LICENSE.txt file for license details.
*/

#include "rumd/RUMD_Error.h"
#include "rumd/SimulationBox.h"


const std::string LE_Error_Code1("simulationBoxSize is too small; it must be at least 8. Change in rumd_base.h and recompile");


//////////////////////////////////////////////////////////
// Rectangular Simulation Box implementation
//////////////////////////////////////////////////////////

RectangularSimulationBox::RectangularSimulationBox() : SimulationBox() {
  AllocateBoxDetails();
  float L = 20.f;

  h_boxDetails[0] = 1; // is cubic box?
  h_boxDetails[1] = h_boxDetails[2] = h_boxDetails[3] = L; 
  h_boxDetails[4] = h_boxDetails[5] = h_boxDetails[6] = 1.f / L;
  
  CopyBoxDetailsToGPU();
}

RectangularSimulationBox::RectangularSimulationBox( float Lx, float Ly, float Lz ) : SimulationBox() {
  AllocateBoxDetails();

  if( Lx == Ly && Ly == Lz )
    h_boxDetails[0] = 1;
  else
    h_boxDetails[0] = 0;
  
  h_boxDetails[1] = Lx; 
  h_boxDetails[2] = Ly; 
  h_boxDetails[3] = Lz;
  h_boxDetails[4] = 1.f / Lx; 
  h_boxDetails[5] = 1.f / Ly; 
  h_boxDetails[6] = 1.f / Lz;
  
  CopyBoxDetailsToGPU();
}

RectangularSimulationBox::~RectangularSimulationBox(){
  cudaFreeHost(h_boxDetails);
  cudaFree(d_boxDetails);
}

SimulationBox* RectangularSimulationBox::MakeCopy(){
  SimulationBox* newBox = new RectangularSimulationBox( h_boxDetails[1], h_boxDetails[2], h_boxDetails[3] );
  return newBox;
}

void RectangularSimulationBox::CopyAnotherBox(const SimulationBox* box_to_copy){
  const RectangularSimulationBox* rect_box_to_copy = dynamic_cast<const RectangularSimulationBox*> (box_to_copy);

  if(rect_box_to_copy){
    float4 box_data = rect_box_to_copy->GetSimulationBox();
    float Lx = box_data.x;
    float Ly = box_data.y;
    float Lz = box_data.z;

    if( Lx == Ly && Ly == Lz )
      h_boxDetails[0] = 1;
    else
      h_boxDetails[0] = 0;
    
    h_boxDetails[1] = Lx; 
    h_boxDetails[2] = Ly; 
    h_boxDetails[3] = Lz;
    h_boxDetails[4] = 1.f / Lx; 
    h_boxDetails[5] = 1.f / Ly; 
    h_boxDetails[6] = 1.f / Lz;
    
    CopyBoxDetailsToGPU();
  }
  else 
    throw RUMD_Error("RectangularSimulationBox", "Copy", "Tried to copy box of a different type");
}

void RectangularSimulationBox::ScaleBox( float factor ){
  h_boxDetails[1] *= factor;
  h_boxDetails[2] *= factor;
  h_boxDetails[3] *= factor;
  h_boxDetails[4] = 1.f / h_boxDetails[1];
  h_boxDetails[5] = 1.f / h_boxDetails[2];
  h_boxDetails[6] = 1.f / h_boxDetails[3];
  CopyBoxDetailsToGPU();  
}

void RectangularSimulationBox::AllocateBoxDetails(){
  // CPU.
  if( cudaMallocHost( (void**) &h_boxDetails, simulationBoxSize * sizeof(float) ) != cudaSuccess )
    throw( RUMD_Error("RectangularSimulationBox","RectangularSimulationBox", "Malloc failed on h_boxDetails" ) );
  
  // GPU.
  if( cudaMalloc( (void**) &d_boxDetails, simulationBoxSize * sizeof(float) ) != cudaSuccess )
    throw( RUMD_Error("RectangularSimulationBox","RectangularSimulationBox","Malloc failed on d_BoxDetails") );
}

void RectangularSimulationBox::CopyBoxDetailsToGPU(){
  cudaMemcpy( d_boxDetails, h_boxDetails, simulationBoxSize * sizeof(float), cudaMemcpyHostToDevice );
  
  if( cudaThreadSynchronize() != cudaSuccess ) 
    throw( RUMD_Error("RectangularSimulationBox","RectangularSimulationBox","CudaMemcpyHostToDevice failed on h_boxDetails => d_boxDetails") );
}

void RectangularSimulationBox::SetBoxDetails( float* details ){
  for(unsigned i=0; i < simulationBoxSize; i++)
    h_boxDetails[i] = details[i];
  
  h_boxDetails[4] = 1.f / h_boxDetails[1]; 
  h_boxDetails[5] = 1.f / h_boxDetails[2]; 
  h_boxDetails[6] = 1.f / h_boxDetails[3];
  
  if( h_boxDetails[1] == h_boxDetails[2] && h_boxDetails[2] == h_boxDetails[3] )
    h_boxDetails[0] = 1;
  else
    h_boxDetails[0] = 0;
  
  CopyBoxDetailsToGPU();
};

void RectangularSimulationBox::ScaleBoxDirection( float factor, int dir ){

  h_boxDetails[dir+1] *= factor;
  h_boxDetails[dir+4] = 1.f / h_boxDetails[dir+1];

  CopyBoxDetailsToGPU();  
}


//////////////////////////////////////////////////////////
// Your Simulation Box implementation
//////////////////////////////////////////////////////////


LeesEdwardsSimulationBox::LeesEdwardsSimulationBox() : RectangularSimulationBox() {
  if(simulationBoxSize < 8)
    throw RUMD_Error("LeesEdwardsSimulationBox","LeesEdwardsSimulationBox",
		     LE_Error_Code1);
  
  // may need to make h_boxDetails double for very slow strain rates
  h_boxDetails[7] = 0.f; 
  CopyBoxDetailsToGPU();
}

LeesEdwardsSimulationBox::LeesEdwardsSimulationBox( float Lx, float Ly, float Lz, float boxShift ) : RectangularSimulationBox(Lx, Ly, Lz){

  if(simulationBoxSize < 8)
    throw RUMD_Error("LeesEdwardsSimulationBox","LeesEdwardsSimulationBox",
		     LE_Error_Code1);
  
  h_boxDetails[7] = boxShift;
  CopyBoxDetailsToGPU();
}


LeesEdwardsSimulationBox::LeesEdwardsSimulationBox(const SimulationBox* box_to_copy) : RectangularSimulationBox() {

  CopyAnotherBox(box_to_copy);
}


SimulationBox* LeesEdwardsSimulationBox::MakeCopy(){
  SimulationBox* newBox = new LeesEdwardsSimulationBox( h_boxDetails[1], h_boxDetails[2], h_boxDetails[3], h_boxDetails[7] );
  return newBox;
}


void LeesEdwardsSimulationBox::CopyAnotherBox(const SimulationBox* box_to_copy){

  RectangularSimulationBox::CopyAnotherBox(box_to_copy);

  const LeesEdwardsSimulationBox* LE_box_to_copy = dynamic_cast<const LeesEdwardsSimulationBox*> (box_to_copy);
  if(LE_box_to_copy){
    h_boxDetails[7] = LE_box_to_copy->GetBoxShift();
    CopyBoxDetailsToGPU();
  }
}

void LeesEdwardsSimulationBox::ScaleBox( float factor )  {
  RectangularSimulationBox::ScaleBox(factor);
  h_boxDetails[7] *= factor;
  CopyBoxDetailsToGPU();  
}

void LeesEdwardsSimulationBox::ScaleBoxDirection( float factor, int dir ){
  RectangularSimulationBox::ScaleBoxDirection(factor, dir);
  if(dir == 1)
    h_boxDetails[7] *= factor;

  CopyBoxDetailsToGPU();  
}

void LeesEdwardsSimulationBox::SetBoxShift( float set_boxShift ) {
  h_boxDetails[7] = set_boxShift;
  CopyBoxDetailsToGPU();
}

void LeesEdwardsSimulationBox::IncrementBoxStrain( float inc_boxStrain ) {
  
  h_boxDetails[7] += inc_boxStrain * h_boxDetails[1];
  h_boxDetails[7] -= h_boxDetails[1] * rintf( h_boxDetails[7]/h_boxDetails[1]);

  CopyBoxDetailsToGPU();
}

